#
# Konstantin Volckov <wolf@virtuozzo.com>
# Denis Silakov <dsilakov@virtuozzo.com>
#
# Copyright (c) 2017-2020 Virtuozzo International GmbH. All rights reserved.
#
# This copyrighted material is made available to anyone wishing to use, modify,
# copy, or redistribute it subject to the terms and conditions of the GNU
# General Public License v.2.  This program is distributed in the hope that it
# will be useful, but WITHOUT ANY WARRANTY expressed or implied, including the
# implied warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  Any Virtuozzo
# trademarks that are incorporated in the source code or documentation are not
# subject to the GNU General Public License and may only be used or replicated
# with the express permission of Virtuozzo.
#
import string
import re

from pykickstart.base import *
from pykickstart.errors import *
from pykickstart.options import *

import gettext
_ = lambda x: gettext.ldgettext("pykickstart", x).decode('utf-8')

import warnings

class PSBM6_PSTORAGE(KickstartCommand):
    removedKeywords = KickstartCommand.removedKeywords
    removedAttrs = KickstartCommand.removedAttrs

    def __init__(self, writePriority=0, *args, **kwargs):
        KickstartCommand.__init__(self, writePriority, *args, **kwargs)
        self.mds_create = []
        self.mds_join = []
        self.cs_join = ""
        self.cs_mpoints_tiers = []
        self.cs_ssd = True
        self.client = ""
        self.client_ssd = True
        self.client_ssd_path_size = []
        self.ipaddr = None
        self.key = ""
        self.clusterpw = ""
        self.reg_timeout = 0   #in minutes.
        self.shaman = True
        self.commandname = "vstorage"

    def __str__(self):
        retval = KickstartCommand.__str__(self)

        if len(self.mds_create) == 2 or (len(self.mds_create) == 1 and self.ipaddr):
            # IP address has been moved to a separate 'ipaddr' option, but the --ip attribute here is kept for compatibility
            ip = self.ipaddr
            if not ip:
                ip = self.mds_create[1]
            retval += "%s mds_create %s --ip=%s\n" % (self.commandname, self.mds_create[0], self.ipaddr)

        if len(self.mds_join) == 2 or (len(self.mds_join) == 1 and self.ipaddr):
            # IP address has been moved to a separate 'ipaddr' option, but the --ip attribute here is kept for compatibility
            ip = self.ipaddr
            if not ip:
                ip = self.mds_join[1]
            retval += "%s mds_join %s --ip=%s\n" % (self.commandname, self.mds_join[0], self.ipaddr)

        if self.cs_join:
            if len(self.cs_mpoints_tiers) == 0:
                retval += "%s cs_join %s" % (self.commandname, self.cs_join)
                if not self.cs_ssd:
                    retval += " --ssd=no"
                retval += "\n"
            else:
                for mpoint, tier, path, size in self.cs_mpoints_tiers:
                    retval += "%s cs_join %s" % (self.commandname, self.cs_join)
                    if not self.cs_ssd:
                        retval += " --mpoint=%s --tier=%s --ssd=no" % (mpoint, tier)
                    else:
                        if path == "" or size == 0:
                            retval += " --mpoint=%s --tier=%s" % (mpoint, tier)
                        else:
                            retval += " --mpoint=%s --tier=%s --ssd=%s,%iM" % (mpoint, tier, path, size)
                    retval += "\n"

        if self.client:
            retval += "%s client %s" % (self.commandname, self.client)
            if not self.client_ssd:
                retval += " --ssd=no"
            else:
                for path, size in self.client_ssd_path_size:
                    if path != "" and size > 0:
                        retval += " --ssd=%s,%iM" % (path, size)
            retval += "\n"
            if not self.shaman:
                retval += "%s shaman off\n" % self.commandname

        if self.ipaddr:
            retval += "%s ipaddr %s\n" % (self.commandname, self.ipaddr)

        if self.key:
            retval += "%s key %s\n" % (self.commandname, self.key)

        if self.clusterpw:
            retval += "%s clusterpw %s\n" % (self.commandname, self.clusterpw)

        if self.reg_timeout > 0:
            retval += "%s reg_timeout %i\n" % (self.commandname, self.reg_timeout)

        return retval

    def validate_clustername(self, cluster):
        if not re.match("^[A-Za-z0-9_-]*$", cluster):
            raise KickstartParseError(_("The specified cluster name contains invalid characters. " \
                "A valid name may contain only the characters a-z, A-Z, 0-9, underscore (_), and dash (-). "), lineno=self.lineno)

    def parse_mds(self, args, pstorage_array):
        if len(args) not in (2, 3):
            raise KickstartParseError(_("Command %s %s requires 1 or 2 arguments") % (self.commandname, args[0]), lineno=self.lineno)
        else:
            # Clustername
            self.validate_clustername(args[1])
            pstorage_array.append(args[1])
            # IP
            if len(args) == 3:
                ip = re.sub("^--ip=", "", args[2])
                if ip:
                    pstorage_array.append(ip)

    def parse_reg_timeout(self, args):
        if len(args) != 2:
            raise KickstartParseError(_("Command %s reg_timeout requires one argument") % self.commandname, lineno=self.lineno)
        else:
            try:
                self.reg_timeout = int(args[1])
            except ValueError:
                raise KickstartParseError(_("Invalid format for the value in the %s reg_timeout command") % self.commandname, lineno=self.lineno)

    def get_arg(self, args, arg):
        for a in args:
            if a.startswith("--%s=" % arg):
                return re.sub("^--%s=" % arg, "", a)

        return ""

    def check_path(self, path):
        return path.startswith("/") and not re.search('[\.]', path)

    def convert_size(self, size_str):
        size = 0
        try:
            if size_str.endswith("M"):
                size = int(size_str.rstrip("M"))
            elif size_str.endswith("G"):
                size = int(size_str.rstrip("G")) * 1024
            elif size_str.endswith("T"):
                size = int(size_str.rstrip("T")) * 1024 * 1024
        except:
            raise KickstartParseError(_("The --ssd size key in the %s cs_join command should be in [0-9][MGT] format") % self.commandname, lineno=self.lineno)

        if size < 0:
            raise KickstartParseError(_("The --ssd size key in the %s cs_join command should be greater than zero") % self.commandname, lineno=self.lineno)

        return size

    def check_ssd_disabled(self, ssd, error=True):
        if ssd != "no":
            if error:
                raise KickstartParseError(_("Invalid ssd for the --ssd key in the %s command, should be \"no\"") % self.commandname, lineno=self.lineno)
            return True
        return False

    def parse_ssd_path_size(self, ssd):
        def format_error():
            raise KickstartParseError(_("The --ssd key in the %s command should be in --ssd=$PATH,$SIZE command") % self.commandname, lineno=self.lineno)
        ssd_path = re.sub(",.*$", "", ssd)
        ssd_size = self.convert_size(re.sub("^.*,", "", ssd))
        if ssd_path:
            if ssd_size == 0:
                format_error()
            if not self.check_path(ssd_path):
                raise KickstartParseError(_("Invalid path for the --ssd key in the %s command") % self.commandname, lineno=self.lineno)
        elif ssd_size != 0:
            format_error()

        return ssd_path, ssd_size

    def parse_cs(self, args):
        self.validate_clustername(args[1])
        self.cs_join = args[1]
        mpoint = self.get_arg(args, "mpoint")
        tier = self.get_arg(args, "tier")
        ssd_path = ""
        ssd_size = 0
        ssd = self.get_arg(args, "ssd")
        if mpoint:
            if ssd:
                if not self.check_ssd_disabled(ssd, error=False):
                    self.cs_ssd = False
                else:
                    ssd_path, ssd_size = self.parse_ssd_path_size(ssd)
            if not self.check_path(mpoint):
                raise KickstartParseError(_("Invalid path for the --mpoint key in the %s cs_join command") % self.commandname, lineno=self.lineno)
            if tier:
                if re.sub("^[0-3]$", "", tier):
                    raise KickstartParseError(_("Invalid tier for the --tier key in the %s cs_join command, should be 0-3") % self.commandname, lineno=self.lineno)
                self.cs_mpoints_tiers.append((mpoint, tier, ssd_path, ssd_size))
            else:
                self.cs_mpoints_tiers.append((mpoint, "0", ssd_path, ssd_size))
            return
        if tier:
            raise KickstartParseError(_("The --tier key in the %s cs_join command can be used only with --mpoint") % self.commandname, lineno=self.lineno)
        if ssd:
            self.cs_ssd = self.check_ssd_disabled(ssd)

    def parse(self, args):
        if len(args) == 0:
            raise KickstartParseError(_("Command %s requires at least one argument") % self.commandname, lineno=self.lineno)
        if len(args) > 5:
            raise KickstartParseError(_("Command %s requires no more than five arguments") % self.commandname, lineno=self.lineno)
        if args[0] == "mds_create":
            self.parse_mds(args, self.mds_create)
        elif args[0] == "mds_join":
            self.parse_mds(args, self.mds_join)
        elif args[0] == "cs_join":
            self.parse_cs(args)
        elif args[0] == "client":
            self.validate_clustername(args[1])
            self.client = args[1]
            ssd = self.get_arg(args, "ssd")
            if ssd:
                if self.check_ssd_disabled(ssd, error=False):
                    ssd_path, ssd_size = self.parse_ssd_path_size(ssd)
                    self.client_ssd_path_size.append((ssd_path, ssd_size))
                self.client_ssd = False
        elif args[0] == "ipaddr":
            self.ipaddr = args[1]
        elif args[0] == "key":
            self.validate_clustername(args[1])
            self.key = args[1]
        elif args[0] == "clusterpw":
            if len(args) != 2:
                raise KickstartParseError(_("Command %s clusterpw requires one argument") % self.commandname, lineno=self.lineno)
            else:
                self.validate_clustername(args[1])
                self.clusterpw = args[1]
        elif args[0] == "reg_timeout":
            self.parse_reg_timeout(args)
        elif args[0] == "shaman":
            if len(args) != 2:
                raise KickstartParseError(_("Command %s requires one argument") % self.commandname, lineno=self.lineno)
            else:
                if args[1] == "off":
                    self.shaman = False
                elif args[1] != "on":
                    raise KickstartParseError(_("The command %s supports only the \"on\" and \"off\" arguments.") % self.commandname, lineno=self.lineno)
        else:
            raise KickstartParseError(_("Unknown command %s") % args[0], lineno=self.lineno)

        # Validation and post-processing (to ensure compatibility)
        for name in ('mds_create', 'mds_join'):
            var = getattr(self, name)
            if len(var) == 1:
                # New style; add IP address to the array to ensure that the old-style code works
                if self.ipaddr:
                    var.append(self.ipaddr)
                else:
                    raise KickstartParseError(_("Command %s %s requires IP address") % (self.commandname, name), lineno=self.lineno)
            elif len(var) == 2:
                # old style; check for inconsistencies and copy the IP address to the new field to ensure the new-style code works
                if self.ipaddr:
                    if self.ipaddr != var[1]:
                        raise KickstartParseError(_("IP address specified in %s %s contradicts to the one in %s ipaddr") % (self.commandname, name, self.commandname), lineno=self.lineno)
                else:
                    self.ipaddr = var[1]
            elif len(var) > 2:
                raise KickstartParseError(_("Internal error! len(self.%s)==%d") % (name, len(var)), lineno=self.lineno)

        if len(self.mds_create) > 0 or len(self.mds_join) > 0 or self.cs_join or self.client:
            # IP address is now common for all 3 components, and required if any of them is selected
            if not self.ipaddr:
                raise KickstartParseError(_("Commands %s mds_create, mds_join, cs_join, client require IP address") % (self.commandname), lineno=self.lineno)

        return self

class VZ7_Vstorage(PSBM6_PSTORAGE):
    def __init__(self, writePriority=100, *args, **kwargs):
        PSBM6_PSTORAGE.__init__(self, writePriority=writePriority, *args, **kwargs)
        self.clustername = ""
        self.productname = "Virtuozzo Storage"
        self.commandname = "vstorage"

    def need_mds_create(self):
        return len(self.mds_create) in (1, 2)

    def need_mds_join(self):
        return len(self.mds_join) in (1, 2)

    def need_mds(self):
        return self.need_mds_create() or self.need_mds_join()

    @property
    def mds_ip(self):
        if self.ipaddr:
            return self.ipaddr
        return self.mds_create[1] if self.need_mds_create() else self.mds_join[1]

    def parse(self, args):
        PSBM6_PSTORAGE.parse(self, args)
        # Sort uniq names
        clusternames = list(set([ name[0] for name in (self.mds_create, self.mds_join) \
            if len(name) == 2 and name[0] ] + [ name for name in \
            (self.cs_join, self.client) if name ]))

        if len(clusternames) == 0:
            return self

        if len(clusternames) > 1:
            raise KickstartParseError(_("Multiplfe different %s cluster " \
                "names were found in kickstart parameters: %s. Only one, unique cluster name is allowed " \
                "per kickstart file." % (self.productname, ", ".join(clusternames))), lineno=self.lineno)

        self.clustername = clusternames[0]

        # Tier and --mpoint options for now moved to part keywork
        if self.cs_mpoints_tiers:
            warnings.warn("The --mpoint and --tier parameters for %s cs_join are not used anymore and will be ignored." % self.commandname)

        return self

class VZ8_Vstorage(VZ7_Vstorage):
    pass
